// TinyPad.cpp : Defines the entry point for the application.
//

#include "stdafx.h"
#include "TinyPad.h"
#include <commctrl.h>
#include <aygshell.h>
#include <sipapi.h>

#include <commdlg.h>
#include <string.h>

#ifndef ChooseFont
#define MYCHOOSEFONT
#endif

#ifdef	MYCHOOSEFONT
#include "fonts.h"
#endif	// MYCHOOSEFONT

#include "find.h"

// Registry definitions

#define REGISTRY_ROOT			TEXT("Software\\yasuho\\TinyPad")
#define VALUE_FONTINFO          TEXT("FontInfo")
#define VALUE_NOWORDWRAP        TEXT("NoWordWrap")

static LOGFONT					s_fontinfo;
static BOOL						s_nowordwrap;

// Commandline arguments

#define ARGS_NEWINST			TEXT("-new")
#define ARGS_NEWINST_LEN		4

// Character code

#define CC_ANSI		0
#define CC_UNICODE	1
#define CC_UTF8		2

typedef struct _CCTAB {
	int code;
	unsigned char sig[8];
	int size;
} CCTAB, *PCCTAB;

static const CCTAB	ccTab[] = {
	{ CC_UNICODE, { 0xff, 0xfe }, 2 },
	{ CC_UTF8, { 0xef, 0xbb, 0xbf }, 3 },
	{ CC_ANSI },
};

static PCCTAB pccTab;
static const PCCTAB pccTabAnsi = (const PCCTAB)&ccTab[2];

// Global Variables:

TCHAR				szTitle[MAX_LOADSTRING];	// The title bar text
TCHAR				szSaveFile[MAX_LOADSTRING];
TCHAR				szNotFound[MAX_LOADSTRING];

// Local Variables:

static TCHAR		szWindowClass[MAX_LOADSTRING];		// The window class name

static HINSTANCE	hInst;		// The current instance
static HWND			hwndCB;		// The command bar handle
static HWND			hMainWnd;	// Main window handle

static SHACTIVATEINFO s_sai;

static HWND			hEdit;

static HFONT		hFont;

static BOOL			isModified;

static TCHAR		FileName[MAX_PATH];

static FINDSTR		fs;

static const TCHAR	szDefaultFilter[] = TEXT("Text (*.txt)\0*.txt\0All (*.*)\0*.*\0");

// Forward declarations of functions included in this code module:

ATOM				MyRegisterClass	(HINSTANCE, LPTSTR);
BOOL				InitInstance	(HINSTANCE, LPTSTR, int);
LRESULT CALLBACK	WndProc			(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK	About			(HWND, UINT, WPARAM, LPARAM);
HWND				CreateRpCommandBar(HWND);
LRESULT CALLBACK	OptionsDlgProc	(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK	SelectWindowDlgProc	(HWND, UINT, WPARAM, LPARAM);


/*
 *	MulDiv
 */
int MulDiv(int m1, int m2, int d)
{
	__int64 x;

	x = (__int64)m1 * (__int64)m2 * 100;
	x = x / d + 50;
	return (int)(x / 100);
}

//
//	InitSettings
//
static void InitSettings()
{
	memset(&s_fontinfo, 0, sizeof s_fontinfo);
	s_nowordwrap = FALSE;

	hFont = (HFONT)GetStockObject(SYSTEM_FONT);
}

//
//	SaveSettings
//
static void SaveSettings()
{
	HKEY hkey = NULL;
	LONG lret;

    lret = RegCreateKeyEx(HKEY_CURRENT_USER, REGISTRY_ROOT, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &hkey, NULL);
    if (lret == ERROR_SUCCESS)
    {
        lret = RegSetValueEx(hkey, VALUE_FONTINFO, 0, REG_BINARY, (LPBYTE)&s_fontinfo, (DWORD)sizeof s_fontinfo);
		if (lret != ERROR_SUCCESS)
        {
	        goto out;
		}

        lret = RegSetValueEx(hkey, VALUE_NOWORDWRAP, 0, REG_DWORD, (LPBYTE)&s_nowordwrap, (DWORD)sizeof s_nowordwrap);
		if (lret != ERROR_SUCCESS)
        {
	        goto out;
		}
	}

out:
	if (hkey) RegCloseKey(hkey);
}

//
//	LoadSettings
//
static void LoadSettings()
{
    LONG lret;
    HKEY hkey = NULL;
    DWORD size;

    lret = RegOpenKeyEx(HKEY_CURRENT_USER, REGISTRY_ROOT, 0, KEY_READ, &hkey);
    if (lret == ERROR_SUCCESS)
    {
        size = sizeof(s_fontinfo);
	    lret = RegQueryValueEx(hkey, VALUE_FONTINFO, 0, NULL, (LPBYTE)&s_fontinfo, &size);
		if (lret != ERROR_SUCCESS)
        {
	        goto out;
		}

        size = sizeof(s_nowordwrap);
	    lret = RegQueryValueEx(hkey, VALUE_NOWORDWRAP, 0, NULL, (LPBYTE)&s_nowordwrap, &size);
		if (lret != ERROR_SUCCESS)
        {
	        goto out;
		}
	}

out:
	if (hkey) RegCloseKey(hkey);
}

//
//	_basename
//
static void _basename(PTCHAR pdst, PTCHAR psrc)
{
	PTCHAR p;

	p = psrc;
	while (*psrc) {
		if (*psrc == L'\\')
			p = &psrc[1];
		psrc++;
	}

	while (*p && *p != L'.')	// omit extension
		*pdst++ = *p++;

	*pdst = L'\0';
}

//
//	GetExtension
//
static PTCHAR GetExtension(PTCHAR pfile)
{
	PTCHAR p;
	PTCHAR px;
	int count;

	count = wcslen(pfile);
	px = NULL;
	p = &pfile[count];
	while (p > pfile) {
		p--;
		if (*p == L'.') {
			px = p;
			break;
		}
	}

	return (px && px[1]) ? px : NULL;
}

//
//	SetTitle
//
static void SetTitle()
{
	TCHAR buf[256];
	TCHAR title[256];

	isModified = SendMessage(hEdit, EM_GETMODIFY, 0, 0);
	_basename(title, FileName);
	wsprintf(buf, TEXT("%s%s%s%s"),
		isModified ? TEXT("*") : TEXT(""),
		title[0] ? title : TEXT(""),
		title[0] ? TEXT(" - ") : TEXT(""),
		szTitle);
	SetWindowText(hMainWnd, buf);
}

//
//	DisacardBuffer
//
static void DiscardBuffer()
{
	// Discard current buffer.

	SendMessage(hEdit, WM_SETTEXT, 0, (LPARAM)TEXT(""));
	SendMessage(hEdit, EM_SETSEL, 0, 0);
	SendMessage(hEdit, EM_SETMODIFY, FALSE, 0);

	memset(FileName, 0, sizeof FileName);
	pccTab = pccTabAnsi;

	SetTitle();
}

//
//	GetOpenName
//
static BOOL GetOpenName(LPTSTR pFileName)
{
	BOOL res;
	OPENFILENAME ofn;
	TCHAR namebuf[MAX_PATH];

	wcscpy(namebuf, pFileName);

	memset(&ofn, 0, sizeof ofn);
	ofn.lStructSize = sizeof ofn;
	ofn.hwndOwner = hEdit;
	ofn.lpstrFile = namebuf;
	ofn.nMaxFile = MAX_PATH;
	ofn.lpstrFilter = szDefaultFilter;

	res = GetOpenFileName(&ofn);
	if (res)
		wcscpy(pFileName, namebuf);

	return res;
}

//
//	GetSaveName
//
static BOOL GetSaveName()
{
	BOOL res;
	OPENFILENAME ofn;
	TCHAR namebuf[MAX_PATH];
	TCHAR filter[MAX_PATH];
	PTCHAR pext;
	PTCHAR pfilter;
	BOOL custom;

	wcscpy(namebuf, FileName);

	memset(&ofn, 0, sizeof ofn);
	ofn.lStructSize = sizeof ofn;
	ofn.hwndOwner = hEdit;
	ofn.lpstrFile = namebuf;
	ofn.nMaxFile = MAX_PATH;

	//ofn.lpstrFilter = TEXT("Text (*.txt)\0*.txt\0All (*.*)\0*.*\0");
	//ofn.lpstrDefExt = TEXT("txt");
	pext = GetExtension(namebuf);
	custom = (pext && _wcsicmp(pext, TEXT(".txt")) != 0);
	pfilter = filter;
	if (custom) {
		pfilter += wsprintf(pfilter, TEXT("%s"), pext);
		*pfilter++ = L'\0';
		pfilter += wsprintf(pfilter, TEXT("%s"), pext);
		*pfilter++ = L'\0';
		ofn.lpstrDefExt = &pext[1];
	} else {
		ofn.lpstrDefExt = TEXT("txt");
	}
	memcpy(pfilter, szDefaultFilter, sizeof szDefaultFilter);
	ofn.lpstrFilter = filter;

	res = GetSaveFileName(&ofn);
	if (res)
		wcscpy(FileName, namebuf);

	return res;
}

//
//	GetCharCodeTable
//
PCCTAB GetCharCodeTable(void *pfile)
{
	PCCTAB pcc;

	for (pcc = (PCCTAB)ccTab; pcc->code != CC_ANSI; pcc++) {
		if (memcmp(pfile, pcc->sig, pcc->size) == 0)
			break;
	}

	return pcc;
}

//
//	FromUnicode
//
static char *FromUnicode(int codepage, PTCHAR punicode)
{
	char *pdst = NULL;
	int count;
	int dstsize;
	int res;

	count = WideCharToMultiByte(codepage, 0, (LPCWSTR)punicode, -1, NULL, 0, NULL, NULL);
	if (!count)
		goto fail;
	dstsize = (count + 1) * sizeof(char);
	pdst = (char *)LocalAlloc(0, dstsize);
	if (!pdst)
		goto fail;
	res = WideCharToMultiByte(codepage, 0, (LPCWSTR)punicode, -1, pdst, dstsize, NULL, NULL);
	if (!res)
		goto fail;

	pdst[count] = 0;

	return pdst;

fail:
	if (pdst) LocalFree(pdst);

	return NULL;
}

//
//	UnicodeToAnsi
//
char *UnicodeToAnsi(PTCHAR punicode)
{
	return FromUnicode(CP_ACP, punicode);
}

//
//	UnicodeToUTF8
//
char *UnicodeToUTF8(PTCHAR punicode)
{
	return FromUnicode(CP_UTF8, punicode);
}

//
//	ToUnicode
//
static PTCHAR ToUnicode(int codepage, char *psrc)
{
	PTCHAR pdst = NULL;
	int count;
	int dstsize;
	int res;

	count = MultiByteToWideChar(codepage, 0, (LPCSTR)psrc, -1, NULL, 0);
	if (!count)
		goto fail;
	dstsize = (count + 1) * sizeof(wchar_t);
	pdst = (PTCHAR )LocalAlloc(0, dstsize);
	if (!pdst)
		goto fail;
	res = MultiByteToWideChar(codepage, 0, (LPCSTR)psrc, -1, pdst, dstsize);
	if (!res)
		goto fail;

	pdst[count] = 0;

	return pdst;

fail:
	if (pdst) LocalFree(pdst);

	return NULL;
}

//
//	AnsiToUnicode
//
PTCHAR AnsiToUnicode(char *pansi)
{
	return ToUnicode(CP_ACP, pansi);
}

//
//	UTF8ToUnicode
//
PTCHAR UTF8ToUnicode(char *putf8)
{
	return ToUnicode(CP_UTF8, putf8);
}

//
//	LoadFile
//
static BOOL LoadFile()
{
	BOOL res = FALSE;
	char *psrc = NULL;
	PTCHAR pdst = NULL;
	HANDLE hfile = NULL;
	DWORD size;
	DWORD nread;
	DWORD err;
	TCHAR tbuf[MAX_PATH];

	// Open the file.
	hfile = CreateFile(FileName, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
	if (!hfile) {
		MessageBox(hEdit, TEXT("CreateFile: failed."), szTitle, MB_OK);
		goto done;
	}
	size = GetFileSize(hfile, NULL);
	if (size == 0xFFFFFFFF) {
		err = GetLastError();
		wsprintf(tbuf, TEXT("GetFileSize: failed. [%x]"), err);
		MessageBox(hEdit, tbuf, szTitle, MB_OK);
		goto done;
	}

	psrc = (char *)LocalAlloc(0, size + sizeof(TCHAR));
	if (!psrc) {
		MessageBox(hEdit, TEXT("No memory."), szTitle, MB_OK);
		goto done;
	}
	if (!ReadFile(hfile, psrc, size, &nread, NULL)) {
		MessageBox(hEdit, TEXT("ReadFile: failed."), szTitle, MB_OK);
		goto done;
	}
	memset(&psrc[size], 0, sizeof(TCHAR)); // null

	// Check unicode text or not.
	pccTab = GetCharCodeTable(psrc);
	switch (pccTab->code) {
	case CC_UNICODE:
		if (!SendMessage(hEdit, WM_SETTEXT, 0, (LPARAM)&psrc[pccTab->size])) {
			MessageBox(hEdit, TEXT("WM_SETTEXT: failed."), szTitle, MB_OK);
			goto done;
		}
		break;
	case CC_UTF8:
		if (!(pdst = UTF8ToUnicode(&psrc[pccTab->size]))) {
			MessageBox(hEdit, TEXT("No memory."), szTitle, MB_OK);
			goto done;
		}
		if (!SendMessage(hEdit, WM_SETTEXT, 0, (LPARAM)pdst)) {
			MessageBox(hEdit, TEXT("WM_SETTEXT: failed."), szTitle, MB_OK);
			goto done;
		}
		break;
	default: // ANSI
		if (!(pdst = AnsiToUnicode(psrc))) {
			MessageBox(hEdit, TEXT("No memory."), szTitle, MB_OK);
			goto done;
		}
		if (!SendMessage(hEdit, WM_SETTEXT, 0, (LPARAM)pdst)) {
			MessageBox(hEdit, TEXT("WM_SETTEXT: failed."), szTitle, MB_OK);
			goto done;
		}
		break;
	}

	SetTitle();

	res = TRUE;

	// TODO: Set file name to window title.

done:
	if (hfile)
		CloseHandle(hfile);
	if (psrc)
		LocalFree(psrc);
	if (pdst)
		LocalFree(pdst);

	return res;
}

//
//	SaveFile
//
static BOOL SaveFile()
{
	BOOL res = FALSE;
	PTCHAR psrc = NULL;
	char *pdst = NULL;
	HANDLE hfile = NULL;
	DWORD size;
	DWORD srcsize;
	DWORD dstsize;
	DWORD wsize;
	int count;

	// Get edit text.
	count = SendMessage(hEdit, WM_GETTEXTLENGTH, 0, 0);
	size = count * sizeof(TCHAR);
	srcsize = size + sizeof(TCHAR);	// null
	psrc = (PTCHAR)LocalAlloc(0, srcsize);
	if (!psrc) {
		MessageBox(hEdit, TEXT("No memory."), szTitle, MB_OK);
		goto done;
	}
	SendMessage(hEdit, WM_GETTEXT, count + 1, (LPARAM)psrc);
	psrc[count] = 0;

	// Check code
	if (!pccTab) pccTab = pccTabAnsi;
	switch (pccTab->code) {
	case CC_UNICODE:
		pdst = (char *)psrc;
		dstsize = size;
		psrc = NULL;
		break;
	case CC_UTF8:
		if (!(pdst = UnicodeToUTF8(psrc))) {
			MessageBox(hEdit, TEXT("No memory."), szTitle, MB_OK);
			goto done;
		}
		dstsize = strlen(pdst);
		break;
	default: // ANSI
		if (!(pdst = UnicodeToAnsi(psrc))) {
			MessageBox(hEdit, TEXT("No memory."), szTitle, MB_OK);
			goto done;
		}
		dstsize = strlen(pdst);
		break;
	}

	// Save to file.
	hfile = CreateFile(FileName, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, 0, 0);
	if (!hfile) {
		MessageBox(hEdit, TEXT("CreateFile: failed."), szTitle, MB_OK);
		goto done;
	}
	if (pccTab->size) {
		if (!WriteFile(hfile, pccTab->sig, pccTab->size, &wsize, NULL)) {
			MessageBox(hEdit, TEXT("WriteFile: failed."), szTitle, MB_OK);
			goto done;
		}
	}
	if (!WriteFile(hfile, pdst, dstsize, &wsize, NULL)) {
		MessageBox(hEdit, TEXT("WriteFile: failed."), szTitle, MB_OK);
		goto done;
	}

	SendMessage(hEdit, EM_SETMODIFY, FALSE, 0);

	SetTitle();

	res = TRUE;

done:
	if (hfile)
		CloseHandle(hfile);
	if (psrc)
		LocalFree(psrc);
	if (pdst)
		LocalFree(pdst);

	return res;
}

//
//	CheckSave
//
static BOOL CheckSave(PTCHAR ptitle)
{
	int res;

	if (SendMessage(hEdit, EM_GETMODIFY, 0, 0)) {
		res = MessageBox(hMainWnd, szSaveFile, ptitle, MB_YESNOCANCEL);
		if (res == IDCANCEL)
			return FALSE;
		else if (res == IDYES) {
			if (!GetSaveName())
				return FALSE;
			if (!SaveFile())
				return FALSE;
		}
	}

	return TRUE;
}

static WNDPROC pEditProc;

//
//	EditSubProc
//
static LRESULT CALLBACK EditSubProc(HWND hWnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	int res;

	switch (msg) {
	case WM_COMMAND:
		return WndProc(hMainWnd, msg, wparam, lparam);
	case WM_LBUTTONDOWN:
		res = WndProc(hMainWnd, msg, wparam, lparam);
		if (res)
			return res;
	}
	return CallWindowProc(pEditProc, hWnd, msg, wparam, lparam);
}

//
//	CopyPathname
//
static void CopyPathname(PTCHAR psrc, PTCHAR pdst)
{
	PTCHAR ps, pe;
	int count;

	count = wcslen(psrc);
	ps = psrc;
	pe = &psrc[count];

	// Find first position
	while (ps < pe && *ps == L'\"')
		ps++;

	// Find last position
	while (pe > ps && pe[-1] == L'\"')
		pe--;

	// Copy actual pathname
	while (ps < pe)
		*pdst++ = *ps++;
	*pdst = L'\0';
}


//
//	WinMain
//
int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPTSTR    lpCmdLine,
					int       nCmdShow)
{
	MSG msg;
	HACCEL hAccelTable;

	// Perform application initialization:
	if (!InitInstance (hInstance, lpCmdLine, nCmdShow)) 
	{
		return FALSE;
	}

	hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_TINYPAD);

	// Main message loop:
	while (GetMessage(&msg, NULL, 0, 0)) 
	{
		if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg)) 
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return msg.wParam;
}

//
//  FUNCTION: MyRegisterClass()
//
//  PURPOSE: Registers the window class.
//
//  COMMENTS:
//
//    It is important to call this function so that the application 
//    will get 'well formed' small icons associated with it.
//
ATOM MyRegisterClass(HINSTANCE hInstance, LPTSTR szWindowClass)
{
	WNDCLASS	wc;

    wc.style			= CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc		= (WNDPROC) WndProc;
    wc.cbClsExtra		= 0;
    wc.cbWndExtra		= 0;
    wc.hInstance		= hInstance;
    wc.hIcon			= LoadIcon(hInstance, MAKEINTRESOURCE(IDI_TINYPAD));
    wc.hCursor			= 0;
    wc.hbrBackground	= (HBRUSH) GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName		= 0;
    wc.lpszClassName	= szWindowClass;

	return RegisterClass(&wc);
}

//
//  FUNCTION: InitInstance(HANDLE, int)
//
//  PURPOSE: Saves instance handle and creates main window
//
//  COMMENTS:
//
//    In this function, we save the instance handle in a global variable and
//    create and display the main program window.
//
BOOL InitInstance(HINSTANCE hInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	RECT	rc;
	LPTSTR	pCmd;
	BOOL	hasFile = FALSE;
	BOOL	fnewInst = FALSE;

	hInst = hInstance;		// Store instance handle in our global variable

	// Initialize global strings
	LoadString(hInstance, IDC_TINYPAD, szWindowClass, MAX_LOADSTRING);
	LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);

	LoadString(hInstance, IDS_SAVEFILE, szSaveFile, MAX_LOADSTRING);
	LoadString(hInstance, IDS_NOTFOUND, szNotFound, MAX_LOADSTRING);

	// load settings from registry

	InitSettings();
	LoadSettings();

	// parse parameters

	pCmd = lpCmdLine;
	if (pCmd)
	{
		if (wcsncmp(pCmd, ARGS_NEWINST, ARGS_NEWINST_LEN) == 0)
		{
			fnewInst = TRUE;

			// skip spaces
			pCmd += ARGS_NEWINST_LEN;
			while (*pCmd && (*pCmd == TEXT(' ') || *pCmd == TEXT('\t')))
				pCmd++;
		}

		hasFile = (pCmd[0] != TEXT('\0'));
	}

	if (!fnewInst)
	{
		//If it is already running, then focus to the active window
		hMainWnd = FindWindow(szWindowClass, NULL);	
		if (hMainWnd)
		{
			// set focus to foremost child window
			// The "| 0x01" is used to bring any owned windows to the foreground and
			// activate them.
			SetForegroundWindow((HWND)((ULONG) hMainWnd | 0x00000001));

			// Sent document name to the active process.
			if (hasFile) {
				COPYDATASTRUCT cds;

				cds.cbData = (wcslen(pCmd) + 1) * sizeof(TCHAR);	// Including nul
				cds.lpData = pCmd;
				SendMessage(hMainWnd, WM_COPYDATA, (WPARAM)hMainWnd, (LPARAM)&cds);
			}

			return 0;
		}
	}
	else
	{
		// create new instance
	}

	MyRegisterClass(hInstance, szWindowClass);
	
	hMainWnd = CreateWindow(szWindowClass, szTitle, WS_VISIBLE,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, hInstance, NULL);
	if (!hMainWnd)
	{	
		return FALSE;
	}

	//When the main window is created using CW_USEDEFAULT the height of the menubar (if one
	// is created is not taken into account). So we resize the window after creating it
	// if a menubar is present
	if (hwndCB)
    {
        RECT rcMenuBar;

		GetWindowRect(hMainWnd, &rc);
        GetWindowRect(hwndCB, &rcMenuBar);
		rc.bottom -= (rcMenuBar.bottom - rcMenuBar.top);
		
		MoveWindow(hMainWnd, rc.left, rc.top, rc.right-rc.left, rc.bottom-rc.top, FALSE);
	}

    GetWindowRect(hMainWnd, &rc);

	if (!(hEdit = CreateWindow(L"edit", NULL,
		WS_CHILD | WS_VISIBLE | ES_MULTILINE | 
		(s_nowordwrap ? (WS_HSCROLL | ES_AUTOHSCROLL) : 0) |
		WS_VSCROLL | ES_AUTOVSCROLL |
		ES_NOHIDESEL,
		0, 0, rc.right - rc.left, rc.bottom - rc.top,
		hMainWnd, NULL, hInstance, NULL))) {
		DestroyWindow(hMainWnd);
		return FALSE;
	}

	pEditProc = (WNDPROC)SetWindowLong(hEdit, GWL_WNDPROC, (LONG)EditSubProc);

	SendMessage(hEdit, EM_LIMITTEXT, -1, 0);	// set to max (=65535 characters)

	// Select new font if it's valid

	if (!s_fontinfo.lfHeight && !s_fontinfo.lfWidth)
	{
		GetObject(hFont, sizeof s_fontinfo, &s_fontinfo);	// Reload
	}
	else
	{
		HFONT hOldFont;

		hOldFont = hFont;
		hFont = CreateFontIndirect(&s_fontinfo);
		SendMessage(hEdit, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(1, 0));
		DeleteObject(hOldFont);
	}

	// Load file if file was associated.
	if (hasFile) {
		CopyPathname(pCmd, FileName);
		(void)LoadFile();
	}

	ShowWindow(hMainWnd, nCmdShow);
	UpdateWindow(hMainWnd);

	SetFocus(hMainWnd);

	return TRUE;
}
//
//  FUNCTION: WndProc(HWND, unsigned, WORD, LONG)
//
//  PURPOSE:  Processes messages for the main window.
//
//  WM_COMMAND	- process the application menu
//  WM_PAINT	- Paint the main window
//  WM_DESTROY	- post a quit message and return
//
//
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HDC hdc;
	int wmId, wmEvent;
	PAINTSTRUCT ps;
	BOOL modify;
	int count;
	TCHAR tbuf[MAX_PATH];
	POINT pt;

	switch (message) 
	{
		case WM_COMMAND:
			wmId    = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 

			// Handle events:
			switch (wmEvent) {
				case EN_UPDATE:
					modify = SendMessage(hEdit, EM_GETMODIFY, 0, 0);
					if (modify != isModified)
						SetTitle();
					break;
			}

			// Parse the menu selections:
			switch (wmId)
			{	
				case IDM_TOOLS_ABOUT:
					DialogBox(hInst, (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
				    break;

				case IDM_FILE_NEW:

					if (!CheckSave(TEXT("New")))
						break;

					DiscardBuffer();

					break;

				case IDM_FILE_OPEN:

					if (!CheckSave(TEXT("Open")))
						break;

					if (!GetOpenName(FileName))
						break;

					if (!LoadFile())
						break;

					break;

				case IDM_FILE_SAVE:

					if (!FileName[0]) {
						if (!GetSaveName())
							break;
					}

					if (!SaveFile())
						break;

					break;

				case IDM_FILE_SAVEAS:

					if (!GetSaveName())
						break;
					
					if (!SaveFile())
						break;

					break;

				case IDM_FILE_EXIT:

					if (!CheckSave(szTitle))
						break;

					SendMessage (hWnd, WM_DESTROY, 0, 0);
					break;

				case IDOK:
					SendMessage (hWnd, WM_CLOSE, 0, 0);
					break;

				case IDM_EDIT_UNDO:
					SendMessage(hEdit, EM_UNDO, 0, 0);
					break;

				case IDM_EDIT_CUT:
					SendMessage(hEdit, WM_CUT, 0, 0);
					break;

				case IDM_EDIT_COPY:
					SendMessage(hEdit, WM_COPY, 0, 0);
					break;

				case IDM_EDIT_PASTE:
					SendMessage(hEdit, WM_PASTE, 0, 0);
					break;

				case IDM_EDIT_SELECTALL:
					SendMessage(hEdit, EM_SETSEL, 0, -1);
					break;

				case IDM_EDIT_PAGEUP:
					GetCaretPos(&pt);
					SendMessage(hEdit, EM_SCROLL, SB_PAGEUP, 0);
					SendMessage(hEdit, WM_LBUTTONDOWN, MK_LBUTTON, MAKELPARAM(pt.x, pt.y + 1));
					SendMessage(hEdit, WM_LBUTTONUP, MK_LBUTTON, MAKELPARAM(pt.x, pt.y + 1));
					break;

				case IDM_EDIT_PAGEDOWN:
					GetCaretPos(&pt);
					SendMessage(hEdit, EM_SCROLL, SB_PAGEDOWN, 0);
					SendMessage(hEdit, WM_LBUTTONDOWN, MK_LBUTTON, MAKELPARAM(pt.x, pt.y + 1));
					SendMessage(hEdit, WM_LBUTTONUP, MK_LBUTTON, MAKELPARAM(pt.x, pt.y + 1));
					break;

				case IDM_EDIT_HOME:
					SendMessage(hEdit, EM_SETSEL, 0, 0);
					SendMessage(hEdit, EM_SCROLLCARET, 0, 0);
					break;

				case IDM_EDIT_END:
					count = SendMessage(hEdit, WM_GETTEXTLENGTH, 0, 0);
					SendMessage(hEdit, EM_SETSEL, count, count);
					SendMessage(hEdit, EM_SCROLLCARET, 0, 0);
					break;

				case IDM_EDIT_FIND:
					fs.hInst = hInst;
					fs.hWnd = hEdit;
					fs.nFindStr = MAX_PATH - 1;

					if (!GetFindStr(&fs))
						break;
					if (fs.nFindStr == 0)
						break;

					// Find the string
					fs.BackSearch = FALSE;
					if (!FindStr(&fs)) {
						wsprintf(tbuf, TEXT("\"%s\"%s"), fs.aFindStr, szNotFound);
						MessageBox(hWnd, tbuf, szTitle, MB_ICONINFORMATION | MB_OK);
					}
					break;
					
				case IDM_EDIT_FINDPREV:
					if (fs.nFindStr == 0)
						break;

					fs.hInst = hInst;
					fs.hWnd = hEdit;
					fs.BackSearch = TRUE;

					// Find the string
					if (!FindStr(&fs)) {
						wsprintf(tbuf, TEXT("\"%s\"%s"), fs.aFindStr, szNotFound);
						MessageBox(hWnd, tbuf, szTitle, MB_ICONINFORMATION | MB_OK);
					}
					break;

				case IDM_EDIT_FINDNEXT:
					if (fs.nFindStr == 0)
						break;

					fs.hInst = hInst;
					fs.hWnd = hEdit;
					fs.BackSearch = FALSE;

					// Find the string
					if (!FindStr(&fs)) {
						wsprintf(tbuf, TEXT("\"%s\"%s"), fs.aFindStr, szNotFound);
						MessageBox(hWnd, tbuf, szTitle, MB_ICONINFORMATION | MB_OK);
					}
					break;

				case IDM_EDIT_REPL:
					fs.hInst = hInst;
					fs.hWnd = hEdit;
					fs.nFindStr = MAX_PATH - 1;
					fs.nReplStr = MAX_PATH - 1;

					if (!GetReplStr(&fs))
						break;
					if (fs.nFindStr == 0 || fs.nReplStr == 0)
						break;

					// Find the string
					fs.BackSearch = FALSE;
					if (!FindStr(&fs)) {
						wsprintf(tbuf, TEXT("\"%s\"%s"), fs.aFindStr, szNotFound);
						MessageBox(hWnd, tbuf, szTitle, MB_ICONINFORMATION | MB_OK);
					}
					break;

				case IDM_EDIT_REPLNEXT:
					if (fs.nFindStr == 0 || fs.nReplStr == 0)
						break;

					fs.hInst = hInst;
					fs.hWnd = hEdit;
					fs.BackSearch = FALSE;

					// Find the string
					if (!ReplStr(&fs)) {
						wsprintf(tbuf, TEXT("\"%s\"%s"), fs.aFindStr, szNotFound);
						MessageBox(hWnd, tbuf, szTitle, MB_ICONINFORMATION | MB_OK);
					}

					// Go to the next string
					SendMessage(hWnd, WM_COMMAND, IDM_EDIT_FINDNEXT, 0);
					break;

				case IDM_TOOLS_FONT: {
					CHOOSEFONT cf;
					HFONT hOldFont;
					LOGFONT lf;

					// Choose font.
					GetObject(hFont, sizeof lf, &lf);
					memset(&cf, 0, sizeof cf);
					cf.lStructSize = sizeof cf;
					cf.hInstance = hInst;
					cf.hwndOwner = hWnd;
					cf.lpLogFont = &lf;
					cf.Flags = CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT | CF_LIMITSIZE;
					cf.nSizeMin = 3;
					cf.nSizeMax = 72;
					if (!ChooseFont(&cf))
						break;

					// Select new font.
					hOldFont = hFont;
					hFont = CreateFontIndirect(&lf);
					SendMessage(hEdit, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(1, 0));
					DeleteObject(hOldFont);

					GetObject(hFont, sizeof s_fontinfo, &s_fontinfo);	// Reload

					// Save settings.
					SaveSettings();
					}
					break;

				case IDM_TOOLS_OPTIONS:
					DialogBox(hInst, (LPCTSTR)IDD_OPTIONS, hWnd, (DLGPROC)OptionsDlgProc);
				    break;

				case IDM_WINDOW_SELECT:
					DialogBox(hInst, (LPCTSTR)IDD_SELECTWINDOW, hWnd, (DLGPROC)SelectWindowDlgProc);
					break;

				case IDM_WINDOW_NEW:
					{
					TCHAR szModule[MAX_PATH];
					PROCESS_INFORMATION	procInfo;

					GetModuleFileName(NULL, szModule, MAX_PATH);
					CreateProcess(szModule, ARGS_NEWINST, NULL, NULL, FALSE, 0, NULL, NULL, NULL, &procInfo);
					}
					break;

				case IDM_WINDOW_NEWFILE:
					{
					TCHAR szParams[MAX_PATH + ARGS_NEWINST_LEN];
					LPTSTR pFileName;
					TCHAR szModule[MAX_PATH];
					PROCESS_INFORMATION	procInfo;

					pFileName = szParams;
					pFileName += wsprintf(pFileName, TEXT("%s "), ARGS_NEWINST);

					if (!GetOpenName(pFileName))
						break;

					GetModuleFileName(NULL, szModule, MAX_PATH);
					CreateProcess(szModule, szParams, NULL, NULL, FALSE, 0, NULL, NULL, NULL, &procInfo);
					}
					break;

				default:
				   return DefWindowProc(hWnd, message, wParam, lParam);
			}
			break;
		case WM_CREATE:
			hwndCB = CreateRpCommandBar(hWnd);
            // Initialize the shell activate info structure
            memset (&s_sai, 0, sizeof (s_sai));
            s_sai.cbSize = sizeof (s_sai);

			pccTab = pccTabAnsi;

			break;
		case WM_PAINT:
			hdc = BeginPaint(hWnd, &ps);
			EndPaint(hWnd, &ps);
			break; 
		case WM_DESTROY:
			DeleteObject(hFont);
			CommandBar_Destroy(hwndCB);
			PostQuitMessage(0);
			break;
		case WM_ACTIVATE:
            // Notify shell of our activate message
			SHHandleWMActivate(hWnd, wParam, lParam, &s_sai, FALSE);
			if (wParam != WA_INACTIVE)
				SetFocus(hEdit);
			else
				SendMessage(hEdit, message, wParam, lParam);
     		break;
		case WM_SETTINGCHANGE:
			SHHandleWMSettingChange(hWnd, wParam, lParam, &s_sai);
     		break;
		case WM_SETFOCUS:
			SetFocus(hEdit);
			break;
		case WM_KILLFOCUS:
			SendMessage(hEdit, message, wParam, lParam);
			break;
		case WM_SIZE:
			MoveWindow(hEdit, 0, 0, LOWORD(lParam), HIWORD(lParam), TRUE);
			break;
		case WM_LBUTTONDOWN:
			{
				SHRGINFO shrg;
				int x, y;
				int cmd;

				x = LOWORD(lParam);
				y = HIWORD(lParam);

				memset(&shrg, 0, sizeof shrg);
				shrg.cbSize = sizeof shrg;
				shrg.hwndClient = hEdit;
				shrg.ptDown.x = x;
				shrg.ptDown.y = y;
				shrg.dwFlags = SHRG_RETURNCMD;
				if (SHRecognizeGesture(&shrg) == GN_CONTEXTMENU) {
					HMENU hMenuPopup;

					// Create popup menu
					hMenuPopup = CreatePopupMenu();

					AppendMenu(hMenuPopup, MF_ENABLED|MF_STRING, IDM_EDIT_UNDO, TEXT("Undo"));
					AppendMenu(hMenuPopup, MF_SEPARATOR, 0, 0);
					AppendMenu(hMenuPopup, MF_ENABLED|MF_STRING, IDM_EDIT_CUT, TEXT("Cut"));
					AppendMenu(hMenuPopup, MF_ENABLED|MF_STRING, IDM_EDIT_COPY, TEXT("Copy"));
					AppendMenu(hMenuPopup, MF_ENABLED|MF_STRING, IDM_EDIT_PASTE, TEXT("Paste"));
					AppendMenu(hMenuPopup, MF_SEPARATOR, 0, 0);
					AppendMenu(hMenuPopup, MF_ENABLED|MF_STRING, IDM_EDIT_SELECTALL, TEXT("Select All"));

					cmd = TrackPopupMenu(hMenuPopup, TPM_LEFTALIGN | TPM_TOPALIGN,
						x, y, 0, hEdit, NULL);

					DestroyMenu(hMenuPopup);

					return cmd;
				}
			}
			break;
		case WM_COPYDATA:
			// Sent from another process.
			PCOPYDATASTRUCT pcds;

			pcds = (PCOPYDATASTRUCT)lParam;

			if (!CheckSave(TEXT("Save")))
				break;

			wcsncpy(FileName, (LPCTSTR)pcds->lpData, MAX_PATH);

			if (!LoadFile())
				break;

			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
   }
   return 0;
}

HWND CreateRpCommandBar(HWND hwnd)
{
	SHMENUBARINFO mbi;

	memset(&mbi, 0, sizeof(SHMENUBARINFO));
	mbi.cbSize     = sizeof(SHMENUBARINFO);
	mbi.hwndParent = hwnd;
	mbi.nToolBarId = IDR_MENU;
	mbi.hInstRes   = hInst;
	mbi.nBmpId     = 0;
	mbi.cBmpImages = 0;

	if (!SHCreateMenuBar(&mbi)) 
		return NULL;

	return mbi.hwndMB;
}

// Mesage handler for the About box.
LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	SHINITDLGINFO shidi;

	switch (message)
	{
		case WM_INITDIALOG:
			// Create a Done button and size it.  
			shidi.dwMask = SHIDIM_FLAGS;
			shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN;
			shidi.hDlg = hDlg;
			SHInitDialog(&shidi);
			return TRUE; 

		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK)
			{
				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;
			}
			break;
	}
    return FALSE;
}

// Mesage handler for the Options dialog box.
LRESULT CALLBACK OptionsDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	SHINITDLGINFO shidi;
	BOOL fdirty;
	BOOL fneedrestart;
	BOOL fnowordwrap;
	int state;

	switch (message)
	{
		case WM_INITDIALOG:
			// word wrap
			state = s_nowordwrap ? BST_CHECKED : BST_UNCHECKED;
			SendDlgItemMessage(hDlg, IDC_NOWORDWRAP, BM_SETCHECK, state, 0);

			// Create a Done button and size it.  
			shidi.dwMask = SHIDIM_FLAGS;
			shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN;
			shidi.hDlg = hDlg;
			SHInitDialog(&shidi);
			return TRUE; 

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
			case IDOK:
				fdirty = FALSE;
				fneedrestart = FALSE;

				// word wrap
				fnowordwrap = IsDlgButtonChecked(hDlg, IDC_NOWORDWRAP) == BST_CHECKED;
				if (fnowordwrap != s_nowordwrap)
				{
					s_nowordwrap = fnowordwrap;

					fneedrestart = TRUE;
					fdirty = TRUE;
				}

				// display restart message
				if (fneedrestart)
					MessageBox(hDlg, TEXT("You will need to restart the application to reflect these change(s)."), szTitle, MB_OK);

				// save settings
				if (fdirty)
					SaveSettings();

				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;

			case IDCANCEL:
				EndDialog(hDlg, (WPARAM)FALSE);
				return TRUE;
			}
			break;
	}
    return FALSE;
}

//
//	add all related windows to the listbox
//
BOOL CALLBACK EnumWindowsProc(HWND hwnd, LPARAM lParam)
{
	HWND hList = (HWND)lParam;
	TCHAR szClass[MAX_PATH];
	TCHAR szTitle[MAX_PATH];
	int index;

	// only taken TinyPad's window
	GetClassName(hwnd, szClass, MAX_PATH);
	if (wcsicmp(szClass, szWindowClass) != 0)
		return TRUE;

	// add to listbox
	GetWindowText(hwnd, szTitle, MAX_PATH);
	index = SendMessage(hList, LB_ADDSTRING, 0, (LPARAM)szTitle);
	if (index == LB_ERR)
		return TRUE;

	// remember window handle
	SendMessage(hList, LB_SETITEMDATA, index, (LPARAM)hwnd);

	// set current selection
	if (hwnd == hMainWnd)
		SendMessage(hList, LB_SETCURSEL, index, 0);

	return TRUE;
}

// Mesage handler for the select window.
LRESULT CALLBACK SelectWindowDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	SHINITDLGINFO shidi;
	HWND hList;
	LONG index;
	HWND hwndFocus;

	hList = GetDlgItem(hDlg, IDC_WINDOW_LIST);

	switch (message)
	{
		case WM_INITDIALOG:
			// Find all window of TinyPad
			EnumWindows(EnumWindowsProc, (LPARAM)hList);

			// Create a Done button and size it.  
			shidi.dwMask = SHIDIM_FLAGS;
			shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN;
			shidi.hDlg = hDlg;
			SHInitDialog(&shidi);
			return TRUE;

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
			case IDC_WINDOW_LIST:
				switch (HIWORD(wParam))
				{
				case LBN_DBLCLK:
					SendMessage(hDlg, WM_COMMAND, IDOK, 0);
					break;
				}
				break;

			case IDOK:
				index = SendMessage(hList, LB_GETCURSEL, 0, 0);
				if (index != LB_ERR)
				{
					hwndFocus = (HWND)SendMessage(hList, LB_GETITEMDATA, index, 0);

					// set focus to foremost child window
					// The "| 0x01" is used to bring any owned windows to the foreground and
					// activate them.
					SetForegroundWindow((HWND)((ULONG) hwndFocus | 0x00000001));
				}

				EndDialog(hDlg, (WPARAM)TRUE);
				return TRUE;

			case IDCANCEL:
				EndDialog(hDlg, (WPARAM)FALSE);
				return TRUE;
			}
			break;
	}
    return FALSE;
}

// vi: ts=4 sw=4
